#if 0
# this file is valid C, Makefile and shell
# this is the only file required to update when changing the version

# https://semver.org/
VERSION_MAJOR=0
VERSION_MINOR=10
VERSION_PATCH=5
#endif

#ifndef JG_VERSION
#define JG_VERSION "0.10.5.1"
#endif
