/* Copyright (c) 2013-2015 Jeffrey Pfau
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
#ifndef HLE_BIOS_H
#define HLE_BIOS_H

#include <mgba-util/common.h>

CXX_GUARD_START

extern const uint8_t hleBios[];

CXX_GUARD_END

#endif
